#!/usr/bin/env python2
# -*- coding: UTF-8 -*-

import traceback
import logging
import os
import re
import ConfigParser
from snack import *

from graphicsmode import XRandr, gmMain

class Logger(object):
    """
    Простейший логгер
    """
    _instance = None
    def __new__(self, *args, **kwargs):
        if not self._instance:
            self._instance = super(Logger, self).__new__(self, *args, **kwargs)
            logging.basicConfig(filename='/linuxcash/logs/current/artix-tools.log',
                    format='%(asctime)s %(levelname)s - %(message)s', level=logging.DEBUG)
        return self._instance

    def info(self, message):
        logging.info(message)

    def error(self, message):
        logging.error(message)

    def warn(self, message):
        logging.warning(message)

# создание логгера
log = Logger()
notAssigned = 'Не назначено'
manual = 'Вручную'

class Option(object):
    """
    Отдельная настройка
    :param name - имя настройки в секции section
    :param text - название настройки; specification - описание настройки
    """
    def __init__(self, name, type, section, text, specification, defaultValue, params='', hidden=False):
        self.name = name
        self.type = type
        self.section = section
        self.text = text
        self.specification = specification
        self.defaultValue = defaultValue
        self.params = params
        self.value = defaultValue
        self.hidden = hidden

    def getValueDescription(self):
        if self.type == 'dict':
            return self.params[self.value]
        return self.value

class ScreenConfiguration(object):
    boolItems = {'True': 'Да', 'False': 'Нет'}
    """
    Настройки экрана
    """
    def __init__(self, section):
        self.rotates = {'normal': 'Отсутствует', 'left': 'Влево', 'right': 'Вправо', 'inverted': 'Инвертировано'}
        self.reflects = {'normal': 'Отсутствует', 'x': 'Горизонтально', 'y': 'Вертикально', 'xy': 'Горизонтально и Вертикально'}

        self.modeOption = Option('mode', 'list', section, 'Разрешение экрана', 'Разрешение экрана', '', list())
        self.interfaceOption = Option('interface', 'list', section, 'Интерфейс',
                'Выбор интерфейса, к которому подключен экран', notAssigned, list())
        self.scaleOption = Option('scale', 'entry', section, 'Изменение размера',
                'Изменение размера выводимого изображения.\nПо умолчанию 1.0x1.0', '1.0x1.0')
        self.reflectOption = Option('reflect', 'dict', section, 'Отражение', 'Отражение по вертикали или горизонтали',
                'normal', self.reflects)
        self.rotateOption = Option('rotate', 'dict', section, 'Поворот экрана', 'Настройка поворота экрана',
                'normal', self.rotates)
        self.gammaOption = Option('gamma', 'entry', section, 'Цветовая гамма',
                'Установка цветовой гаммы.\nПо умолчанию 1.0:1.0:1.0', '1.0:1.0:1.0')

        self.sameAs = Option('same-as', 'list', section, 'Дублирование', 'Дублирование выбранного монитора', notAssigned, list())
        self.leftOf = Option('left-of', 'list', section, 'Слева от', 'Слева от выбранного монитора', notAssigned, list())
        self.rightOf = Option('right-of', 'list', section, 'Справа от', 'Справа от выбранного монитора', notAssigned, list())
        self.above = Option('above', 'list', section, 'Сверху от', 'Сверху от выбранного монитора', notAssigned, list())
        self.below = Option('below', 'list', section, 'Снизу от', 'Снизу от выбранного монитора', notAssigned, list())
        self.relativePositionGroup = [self.sameAs, self.leftOf, self.rightOf, self.above, self.below]

        self.primary = Option('primary', 'list', section, 'Основной', 'Является ли монитор основным', 'False', self.boolItems)
        self.application = Option('application', 'listWithEntry', section, 'Приложение', 'Приложение для запуска', notAssigned, [notAssigned, '/usr/bin/cutecom', manual])

        self.multipleScreenOptions = self.relativePositionGroup + [self.primary, self.application]

        self.options = [self.modeOption, self.scaleOption, self.reflectOption, self.rotateOption, self.gammaOption] + self.multipleScreenOptions

        # при смене интерфейса экрана необходимо удалить секцию, к которой настройки относились изначально
        self.initSection = section

    def getOptions(self):
        """
        Получения списка опций экрана
        """
        return [option for option in self.options if not option.hidden]

    def setModes(self, modes):
        """
        Установка доступных для экрана режимов
        """
        self.modeOption.params = modes

    def setInterfaces(self, interfaces):
        """
        Установка доступных для настройки интерфейсов
        """
        self.interfaceOption.params = interfaces

    def setUnusedInterfaces(self, interfaces):
        """
        Установка значений неиспользуемых интерфейсов для относительной позиции мониторов
        """
        for option in self.relativePositionGroup:
            option.params = interfaces

    def setSection(self, section):
        """
        Установка секции для настроек экрана
        """
        for option in self.getOptions():
            option.section = section

    def getAnnotation(self):
        """
        Строка для отображения в списке настроек
        """
        if self.interfaceOption.value == notAssigned:
            return "Не сконфигурирован"
        return "%-6s %s" % (self.interfaceOption.value, self.modeOption.value)

    def hideMultipleScreenOptions(self):
        for option in self.multipleScreenOptions:
            option.value = option.defaultValue
            option.hidden = True

    def showMultipleScreenOptions(self):
        for option in self.multipleScreenOptions:
            option.hidden = False

    def setPrimary(self, value):
        self.primary.value = value
        self.application.hidden = True if value == 'True' else False

class Config(object):
    """
    Предоставляет настройки из ini файла
    """
    def __init__(self):
        self.iniFile = "/linuxcash/cash/conf/graphicsmode.ini"
        self.config = ConfigParser.RawConfigParser()

        # открытие конфига
        if os.path.isfile(self.iniFile):
            self.config.read(self.iniFile)
            self.checkMainSection()
        else:
            self.checkMainSection()
            self.write()

    def getBool(self, section, option, defaultValue):
        value = defaultValue
        try:
            value = self.config.getboolean(section, option)
        except:
            pass
        return value

    def getString(self, section, option, defaultValue):
        value = defaultValue
        try:
            value = self.config.get(section, option)
        except:
            pass
        return value

    def getOption(self, option):
        try:
            value = self.config.get(option.section, option.name)
        except:
            value = option.defaultValue
        return value

    def sections(self):
        return self.config.sections()

    def set(self, section, option, value):
        self.config.set(section, option, value)

    def setOption(self, option, value):
        if not value:
            self.config.remove_option(option.section, option.name)
        else:
            self.config.set(option.section, option.name, value)

    def checkMainSection(self):
        """
        Функция для проверки главной секции в конфиге
        """
        if not self.config.has_section('main'):
            self.config.add_section('main')
        if self.getBool('main', 'setMode', False):
            self.config.set('main', 'setMode', 'True')
        else:
            self.config.set('main', 'setMode', 'False')

    def checkSection(self, section):
        """
        Функция для проверки секции в конфиге
        """
        if not self.config.has_section(section):
            self.config.add_section(section)

    def removeSection(self, section):
        self.config.remove_section(section)

    def setScreenConfiguration(self, cfg):
        if cfg.initSection != notAssigned:
            self.removeSection(cfg.initSection)
        section = cfg.interfaceOption.value
        if section == notAssigned:
            return
        self.checkSection(section)
        cfg.initSection = section
        for option in cfg.getOptions():
            if option.value and option.value != notAssigned:
                self.setOption(option, option.value)

    def removeScreenSections(self):
        """
        Удаление всех секций с настройками экранов
        """
        sections = self.sections()
        sections.remove('main')
        for section in sections:
            self.removeSection(section)

    def removeLastScreenSection(self):
        """
        Удаление секции настроек второго экрана
        """
        sections = self.sections()
        sections.remove('main')
        if len(sections) == 2:
            self.removeSection(sections[1])

    def getUsedInterfaces(self):
        """
        Список используемых интерфейсов
        """
        return [screen.interfaceOption.value for screen in self.getScreenConfigurations()]

    def getScreenConfigurations(self):
        sections = self.sections()
        sections.remove('main')
        result = list()

        screens = 0
        primaryFound = False
        for screenSection in sections:
            if screens == 2:
                break
            screenConfiguration = ScreenConfiguration(screenSection)
            screenConfiguration.interfaceOption.value = screenSection

            for option in screenConfiguration.getOptions():
                option.value = self.getOption(option)

            result.append(screenConfiguration)
            primaryFound = (screenConfiguration.primary.value == 'True')
            screens = screens + 1

        # миграция настроек
        if screens and not self.getString('main', 'screens', ''):
            self.set('main', 'screens', 'one')

        # есть несколько ситуаций:
        # 1. primary указан в настройках, но они ещё не применились - это и будет primary-монитор
        # 2. primary не указан в настройках, однако уже есть primary-монитор - это и будет primary-монитор
        # 3. primary не указан в настройках и нет primary-монитора. Тогда полагаем, что первый в списке от xrandr и будет primary
        if not primaryFound:
            output = XRandr().getPrimaryOutput()
            for screenConfiguration in result:
                if screenConfiguration.interfaceOption.value == output:
                    screenConfiguration.setPrimary('True')
                else:
                    screenConfiguration.setPrimary('False')
        return result

    def write(self):
        """
        Запись конфигурации в файл
        """
        with open(self.iniFile, 'wb') as configFile:
            self.config.write(configFile)

class MainForm(object):
    screensItems = {'all': 'Все', 'one': '1', 'two': '2'}
    setMode = Option('setMode', 'list', 'main', 'Изменять настройки', 'Если выключено, используется автоматическая настройка.',
            'False', ScreenConfiguration.boolItems)
    screens = Option('screens', 'list', 'main', 'Используемые экраны', 'Количество используемых экранов',
            'all', screensItems)
    screen1 = Option('screen', 'screen', str(), 'Экран 1', 'Настройки определенного экрана.', 'Не сконфигурирован')
    screen2 = Option('screen', 'screen', str(), 'Экран 2', 'Настройки определенного экрана.', 'Не сконфигурирован')

    def __init__(self):
        # создание экрана
        self.screen = SnackScreen()
        self.screen.setColor('ROOT', 'white', 'brightmagenta')
        self.screen.setColor('HELPLINE', 'white', 'brightmagenta')

        self.config = Config()
        self.xRandr = XRandr()

        # проверяем подлючены ли по SSH!
        self.ssh = True if 'SSH_TTY' in os.environ else False

        if self.ssh:
            self.bb = ButtonBar(self.screen, (('Применить (A)', 'apply'), ("Сохранить (S)", "save"), ("Выход (ESC)", "exit")), compact = 1)
        else:
            self.bb = ButtonBar(self.screen, (("Сохранить (S)", "save"), ("Выход (ESC)", "exit")), compact = 1)

        self.screenConfigurations = self.config.getScreenConfigurations()
        while len(self.screenConfigurations) < 2:
            self.screenConfigurations.append(ScreenConfiguration(notAssigned))

        # текст бокс для описаний опции
        self.textBox = Textbox(50, 4, '', wrap = 1)

        # список с настройками
        self.optionsList = Listbox(10, width = 50, returnExit = 1)
        self.optionsList.setCallback(callBack, [self.optionsList, self.textBox])

    def start(self):
        try:
            log.info(u"Получаем список режимов")
            modes, interfaces = self.xRandr.getModes()
        except Exception, e:
            log.error(u"Не удалось получить список режимов: %s" % repr(e))
            self.messageBox('Не удалось получить список режимов', str(e), (("ОК", "ok"),))

        mode = self.config.getString('main', 'mode', '800x600')
        if mode not in modes:
            self.config.set('main', 'mode', '' if not modes else modes[0])

        g = GridForm(self.screen, "Настройки экранов", 1, 4)
        g.add(self.optionsList, 0, 1)
        g.add(self.textBox, 0, 2)
        g.add(self.bb, 0, 3)

        # добавление горячих клавиш
        g.addHotKey("ESC")
        g.addHotKey('s')
        g.addHotKey('a')
        current = ''

        while True:
            self.__createOptionsList(modes)

            try:
                self.optionsList.setCurrent(current)
            except Exception, e:
                pass

            self.textBox.setText(self.optionsList.current().specification)
            # отображаем форму
            result = g.run()
            current = self.optionsList.current()
            resultBtn = self.bb.buttonPressed(result)

            # обработка результатов выполнения
            if result == self.optionsList:
                if current == self.screens:
                    val = listDlg(self.screen, current.params, self.config.getOption(current), current.text)
                    # если настраиваются все мониторы - необходимо очистить конфигурацию от индивидуальных настроек
                    if val == 'all':
                        self.screenConfigurations[0] = ScreenConfiguration(notAssigned)
                        self.screenConfigurations[1] = ScreenConfiguration(notAssigned)
                        self.config.removeScreenSections()
                    elif val == 'one':
                        self.screenConfigurations[0].hideMultipleScreenOptions()
                        self.screenConfigurations[1] = ScreenConfiguration(notAssigned)
                        self.config.removeLastScreenSection()
                    else:
                        for screenConfiguration in self.screenConfigurations:
                            screenConfiguration.showMultipleScreenOptions()
                    self.config.setOption(current, val)
                elif current.type == 'list':
                    val = listDlg(self.screen, current.params, self.config.getOption(current), current.text)
                    self.config.setOption(current, val)
                elif current.type == 'entry':
                    val = entryDlg(self.screen, self.config.getOption(current), current.text)
                    self.config.setOption(current, val)
                elif current == self.screen1:
                    self.config.setScreenConfiguration(self.screenConfigurationDlg(self.screenConfigurations[0]))
                elif current == self.screen2:
                    self.config.setScreenConfiguration(self.screenConfigurationDlg(self.screenConfigurations[1]))
            elif resultBtn == 'save' or result == "s":
                self.__saveConfig()
            elif (resultBtn == 'apply' or result == "a") and self.ssh:
                self.__applyConfig()
            elif resultBtn == 'exit' or result == "ESC" :
                break

    def finish(self):
        if self.screen:
            self.screen.finish()

    def screenConfigurationDlg(self, cfg):
        """
        Функция показа диалогового окна с полем ввода
        :param cfg - конфигурация экрана ScreenConfiguration
        """
        buttons = ButtonBar(self.screen, [("Выход", "exit")], compact=1)
        textBox = Textbox(48, 3, '', wrap=1)
        form = GridForm(self.screen, "Настройки экрана 1", 20, 65)

        options = Listbox(len(cfg.getOptions()) + 2, width=48, returnExit=1)
        options.setCallback(callBack, [self.optionsList, textBox])

        form.addHotKey("ESC")
        form.add(options, 1, 1)
        form.add(textBox, 1, 2)
        form.add(buttons, 1, 3)

        modesTmp, allInterfaces = self.xRandr.getModes()
        modes = self.xRandr.getInterfacesModes()
        options.setCallback(callBack, [options, textBox])
        selectedOption = ''

        textBox.setText(cfg.interfaceOption.specification)

        availInterfaces = [notAssigned] + [i for i in allInterfaces if i not in self.config.getUsedInterfaces()]
        if cfg.interfaceOption.value not in availInterfaces:
            availInterfaces.append(cfg.interfaceOption.value)
        cfg.setInterfaces(availInterfaces)

        cfg.setUnusedInterfaces([notAssigned] + [interface for interface in self.config.getUsedInterfaces() if interface != cfg.interfaceOption.value])

        while True:
            options.clear()
            options.append(formatListLine(cfg.interfaceOption.text, cfg.interfaceOption.value), cfg.interfaceOption)

            if cfg.interfaceOption.value != notAssigned:
                cfg.setModes(modes[cfg.interfaceOption.value])
                cfg.setSection(cfg.interfaceOption.value)
                if not cfg.modeOption.value:
                    cfg.modeOption.value = modes[cfg.interfaceOption.value][0] if modes[cfg.interfaceOption.value] else ''

                for option in cfg.getOptions():
                    if option.name == 'primary': # TODO: костыль. Нужно переделать!
                        options.append(formatListLine(option.text, ScreenConfiguration.boolItems[option.value]), option)
                    else:
                        options.append(formatListLine(option.text, option.getValueDescription()), option)
            try:
                options.setCurrent(selectedOption)
            except Exception, e:
                pass

            result = form.run()
            selectedOption = options.current()

            if result == options:
                if selectedOption.type == 'list' or selectedOption.type == 'dict':
                    selectedOption.value = listDlg(self.screen, selectedOption.params, selectedOption.value, selectedOption.text)
                elif selectedOption.type == 'entry':
                    selectedOption.value = entryDlg(self.screen, selectedOption.value, selectedOption.text)
                elif selectedOption.type == 'listWithEntry':
                    selectedOption.value = listWithEntryDlg(self.screen, selectedOption.params, selectedOption.value, selectedOption.text)
                self.__clearRelativePositions(cfg, selectedOption)
                self.__clearPrimary(cfg, selectedOption)
            elif buttons.buttonPressed(result) == 'exit' or result == "ESC":
                self.screen.popWindow()
                return cfg

    def __clearRelativePositions(self, cfg, selectedOption):
        for option in cfg.relativePositionGroup:
            if option.value == selectedOption.value and option.name != selectedOption.name:
                option.value = notAssigned

    def __clearPrimary(self, cfg, selectedOption):
        if selectedOption.name != 'primary':
            return
        # если выбранная настройка primary и она была выключена, то ищем первый попавшийся отличный от текущего монитор и устанавливаем его в primary
        if selectedOption.value == 'False':
            cfg.setPrimary('False')
            for screenConfiguration in self.screenConfigurations:
                if cfg.interfaceOption.value != screenConfiguration.interfaceOption.value:
                    screenConfiguration.setPrimary('True')
                    break
        else: # иначе настройка установлена в True и необходимо сбросить настройки у остальных мониторов
            cfg.setPrimary('True')
            for screenConfiguration in self.screenConfigurations:
                if cfg.interfaceOption.value != screenConfiguration.interfaceOption.value:
                    screenConfiguration.setPrimary('False')

    def __createOptionsList(self, modes):
        '''
        Очистка и заполнение self.optionsList опциями
        '''
        mode = self.config.getString('main', 'mode', '800x600')
        modeOption = Option('mode', 'list', 'main', 'Разрешение экрана', 'Разрешение для всех мониторов', mode if mode in modes else '', modes)
        fbOption = Option('fb', 'entry', 'main', 'Размер скрина',
                'Переконфигурация экрана до указанного размера.\nПо умолчанию равно разрешению экрана', '')

        self.optionsList.clear()
        self.optionsList.append(formatListLine(self.setMode.text, ScreenConfiguration.boolItems[self.config.getOption(self.setMode)]), self.setMode)

        screensOption = self.config.getOption(self.screens)
        if self.config.getOption(self.setMode) == 'True':
            self.optionsList.append(formatListLine(self.screens.text, self.screensItems[self.config.getOption(self.screens)]), self.screens)
            if screensOption == 'one' or screensOption == 'two':
                self.optionsList.append(formatListLine("Экран 1", self.screenConfigurations[0].getAnnotation()), self.screen1)
                if screensOption == 'two':
                    self.optionsList.append(formatListLine("Экран 2", self.screenConfigurations[1].getAnnotation()), self.screen2)
            else:
                self.optionsList.append(formatListLine(modeOption.text, self.config.getOption(modeOption)), modeOption)
            self.optionsList.append(formatListLine(fbOption.text, self.config.getOption(fbOption)), fbOption)

    def __saveConfig(self):
        try:
            self.config.write()
        except Exception, e:
            log.error(u"Ошибка сохранения: %s" % str(e))
            self.messageBox('Ошибка сохранения', str(e), (("ОК", "ok"),))
        else:
            log.info(u"Конфигурационный файл успешно сохранен")
            self.messageBox('Сохранено', 'Конфигурационный файл успешно сохранен!', (("ОК", "ok"),))

    def __applyConfig(self):
        try:
            self.config.write()
            gmMain()
        except Exception, e:
            log.error(u"Ошибка: %s" % str(e))
            self.messageBox('Ошибка', str(e), (("ОК", "ok"),))
        else:
            log.info(u"Настройки применены")
            self.messageBox('Выполнено', 'Настройки применены!', (("ОК", "ok"),))

    def messageBox(self, title, text, buttons):
        '''
        Функция показа диалогового окна с кнопками
        '''
        bb = ButtonBar(self.screen, buttons, compact = 1)
        g = GridForm(self.screen, title, 1, 2)
        g.add(Label(text), 0, 0, (0, 0, 0, 1))
        g.add(bb, 0, 1)
        return bb.buttonPressed(g.runOnce())


def listDlg(screen, params, current, title):
    """
    Функция показа диалогового окна со списком
    """
    l = len(params)
    height = l if l < 6 else 6
    scroll = 1 if l > 6 else 0
    li = Listbox(height = height, width = 28, scroll = scroll, returnExit = 1)
    if type(params) == list:
        for param in params:
            li.append(param, param)
    elif type(params) == dict:
        for key in params:
            li.append(params[key], key)

    g = GridForm(screen, title, 1, 1)
    g.add(li, 0, 0)
    g.addHotKey("ESC")

    try:
        li.setCurrent(current)
    except Exception, e:
        pass

    return li.current() if g.runOnce() == li else current

def entryDlg(screen, current, title):
    """
    Функция показа диалогового окна с полем ввода
    """
    entry = Entry(22)
    entry.set(current)
    bb = ButtonBar(screen, (("Применить", "ok"), ("Отмена", "canсel")), compact = 1)
    g = GridForm(screen, title, 1, 2)
    g.add(entry, 0, 0, (0, 0, 0, 1))
    g.add(bb, 0, 1)
    g.addHotKey("ESC")
    g.addHotKey("ENTER")

    result = g.runOnce()
    return entry.value() if bb.buttonPressed(result) == 'ok' or result == "ENTER" else current

def listWithEntryDlg(screen, params, current, title):
    '''
    Основная проблема этого метода заключается в том, что добавляемое вручную значение хардкодно зависит от длины списка
    '''
    if current not in params:
        params.insert(2, current)
    l = len(params)
    height = l if l < 6 else 6
    scroll = 1 if l > 6 else 0
    li = Listbox(height = height, width = 28, scroll = scroll, returnExit = 1)
    for param in params:
        li.append(param, param)

    g = GridForm(screen, title, 1, 1)
    g.add(li, 0, 0)
    g.addHotKey("ESC")
    g.addHotKey("ENTER")

    try:
        li.setCurrent(current)
    except Exception, e:
        pass

    runResult = g.runOnce()

    if runResult == "ENTER":
        if li.current() == manual:
            result = entryDlg(screen, params[2] if len(params) > 3 else '', title).strip()
            if len(result):
                if len(params) == 3:
                    params.insert(2, result)
                else:
                    params[2] = result
                return result
            else:
                return current
        else:
            return li.current()
    else:
        return current

def callBack(obj):
    """
    функция для показа справки по определнному полю в списке настроек
    """
    li = obj[0]
    textBox = obj[1]
    textBox.setText("\n\n\n\n")
    textBox.setText(li.current().specification)

def formatListLine(text, value):
    l = len(text)
    l = l / 2 + (l & 1)
    return ' ' + text + ':' + (' ' * (20 - l)) + value

if __name__ == "__main__":
    try:
        m = MainForm()
        m.start()
    except:
        log.error(traceback.format_exc())
    finally:
        m.finish()
